/*
 * EpsilonClosure.h
 *
 *  Created on: 23. 3. 2014
 *	  Author: Tomas Pecka
 */

#ifndef EPSILON_CLOSURE_H_
#define EPSILON_CLOSURE_H_

#include <common/multipleDispatch.hpp>
#include <algorithm>
#include <deque>
#include <set>

#include "automaton/Automaton.h"
#include <automaton/common/State.h>

namespace automaton {

namespace properties {

class EpsilonClosure : public std::SingleDispatchLastStaticParam<std::set<automaton::State>, automaton::AutomatonBase, const automaton::State&> {
public:
	static std::set<automaton::State> epsilonClosure( const automaton::Automaton & automaton, const automaton::State & state );

	/**
	 * Computes epsilon closure of a state in epsilon nonfree automaton
	 */
	static std::set<automaton::State> epsilonClosure( const automaton::EpsilonNFA & fsm, const automaton::State & state );
	static std::set<automaton::State> epsilonClosure( const automaton::MultiInitialStateNFA & fsm, const automaton::State & state );
	static std::set<automaton::State> epsilonClosure( const automaton::NFA & fsm, const automaton::State & state );
	static std::set<automaton::State> epsilonClosure( const automaton::DFA & fsm, const automaton::State & state );
	static std::set<automaton::State> epsilonClosure( const automaton::ExtendedNFA & fsm, const automaton::State & state );
	static std::set<automaton::State> epsilonClosure( const automaton::CompactNFA & fsm, const automaton::State & state );

	static EpsilonClosure& getInstance() {
		static EpsilonClosure res;
		return res;
	}
};

} /* namespace properties */

} /* namespace automaton */

#endif /* EPSILON_CLOSURE_H_ */
