/*
 * SingleInitialState.cpp
 *
 *  Created on: 20. 9. 2014
 *	  Author: Tomas Pecka
 */

#include "SingleInitialState.h"

#include <algorithm>
#include <set>

#include <exception/AlibException.h>
#include <automaton/FSM/DFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/MultiInitialStateNFA.h>
#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/CompactNFA.h>

#include <label/LabelSetLabel.h>

namespace automaton {

namespace simplify {

automaton::Automaton SingleInitialState::convert(const Automaton& automaton) {
	return getInstance().dispatch(automaton.getData());
}

automaton::NFA SingleInitialState::convert(const automaton::MultiInitialStateNFA& automaton) {
	// step 1, 3
	std::set<label::Label> labelSet;
	for(const auto& q : automaton.getInitialStates()) {
		labelSet.insert(q.getName());
	}
	automaton::State q0(label::Label(label::LabelSetLabel(std::move(labelSet))));

	automaton::NFA res(q0);
	res.setInputAlphabet(automaton.getInputAlphabet());
	for(const auto& q : automaton.getStates())
		res.addState(q);

	// step 2
	for(const auto& q : automaton.getInitialStates()) {
		for(const auto& kv: automaton.getTransitionsFromState(q)) {
			for(const auto& to : kv.second) {
				res.addTransition(q0, kv.first.second, to);
			}
		}
	}

	for(const auto& t : automaton.getTransitions()) {
		for(const auto& q : t.second) {
			res.addTransition(t.first.first, t.first.second, q);
		 }
	}

	// step 4, 5
	std::set<automaton::State> intersection;
	std::set_intersection(res.getFinalStates().begin(), res.getFinalStates().end(), automaton.getInitialStates().begin(), automaton.getInitialStates().end(), std::inserter(intersection, intersection.begin()));

	res.setFinalStates(automaton.getFinalStates());
	if(intersection.size() != 0) {
		res.addFinalState(q0);
	}

	return res;
}

auto SingleInitialStateMultiInitialStateNFA = SingleInitialState::RegistratorWrapper<automaton::NFA, automaton::MultiInitialStateNFA>(SingleInitialState::getInstance(), SingleInitialState::convert);

automaton::DFA SingleInitialState::convert(const automaton::DFA& automaton) {
	return automaton;
}

auto SingleInitialStateDFA = SingleInitialState::RegistratorWrapper<automaton::DFA, automaton::DFA>(SingleInitialState::getInstance(), SingleInitialState::convert);

automaton::EpsilonNFA SingleInitialState::convert(const automaton::EpsilonNFA& automaton) {
	return automaton;
}

auto SingleInitialStateEpsilonNFA = SingleInitialState::RegistratorWrapper<automaton::EpsilonNFA, automaton::EpsilonNFA>(SingleInitialState::getInstance(), SingleInitialState::convert);

automaton::NFA SingleInitialState::convert(const automaton::NFA& automaton) {
	return automaton;
}

auto SingleInitialStateNFA = SingleInitialState::RegistratorWrapper<automaton::NFA, automaton::NFA>(SingleInitialState::getInstance(), SingleInitialState::convert);

automaton::CompactNFA SingleInitialState::convert(const automaton::CompactNFA& automaton) {
	return automaton;
}

auto SingleInitialStateCompactNFA = SingleInitialState::RegistratorWrapper<automaton::CompactNFA, automaton::CompactNFA>(SingleInitialState::getInstance(), SingleInitialState::convert);

automaton::ExtendedNFA SingleInitialState::convert(const automaton::ExtendedNFA& automaton) {
	return automaton;
}

auto SingleInitialStateExtendedNFA = SingleInitialState::RegistratorWrapper<automaton::ExtendedNFA, automaton::ExtendedNFA>(SingleInitialState::getInstance(), SingleInitialState::convert);

} /* namespace simplify */

} /* namespace automaton */
