/*
 * SingleInitialState.h
 *
 *  Created on: 20. 9. 2014
 *	  Author: Tomas Pecka
 */

#ifndef _SINGLE_INITIAL_STATE_H_
#define _SINGLE_INITIAL_STATE_H_

#include <common/multipleDispatch.hpp>
#include <automaton/Automaton.h>

namespace automaton {

namespace simplify {

/**
 * Makes finite automaton's transition function convert.
 * Source: Melichar: Algorithm 2.22
 */
class SingleInitialState : public std::SingleDispatch<automaton::Automaton, automaton::AutomatonBase> {
public:
	/**
	 * Computes epsilon closure of a state in epsilon nonfree automaton
	 */
	static automaton::Automaton convert(const automaton::Automaton& automaton);

	static automaton::NFA convert(const automaton::MultiInitialStateNFA& automaton);

	static automaton::DFA convert(const automaton::DFA& automaton);
	static automaton::NFA convert(const automaton::NFA& automaton);
	static automaton::EpsilonNFA convert(const automaton::EpsilonNFA& automaton);
	static automaton::ExtendedNFA convert(const automaton::ExtendedNFA& automaton);
	static automaton::CompactNFA convert(const automaton::CompactNFA& automaton);

	static SingleInitialState& getInstance() {
		static SingleInitialState res;
		return res;
	}
};

} /* namespace simplify */

} /* namespace automaton */

#endif /* _SINGLE_INITIAL_STATE_H_ */
