/*
 * AutomataConcatenationEpsilonTransition.cpp
 *
 *  Created on: 20. 11. 2014
 *	  Author: Tomas Pecka
 */

#include "AutomataConcatenationEpsilonTransition.h"
#include <exception/AlibException.h>
#include <label/Label.h>
#include "common/PairLabel.h"

#define AUTOMATON_FIRST  1
#define AUTOMATON_SECOND 2

namespace automaton {

namespace transform {

automaton::Automaton AutomataConcatenationEpsilonTransition::concatenation(const automaton::Automaton& first, const automaton::Automaton& second) {
	return getInstance().dispatch(first.getData(), second.getData());
}

automaton::EpsilonNFA AutomataConcatenationEpsilonTransition::concatenation(const automaton::DFA& first, const automaton::DFA& second) {
	automaton::EpsilonNFA res(pairLabel(label::labelFrom(AUTOMATON_FIRST), first.getInitialState()));

	for(const auto& symbol : first.getInputAlphabet())
		res.addInputSymbol(symbol);
	for(const auto& symbol : second.getInputAlphabet())
		res.addInputSymbol(symbol);

	for(const auto& q : first.getStates())
		res.addState(pairLabel(label::labelFrom(AUTOMATON_FIRST), q));
	for(const auto& q : second.getStates())
		res.addState(pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	for(const auto& q : second.getFinalStates())
		res.addFinalState(pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	for(const auto& t : first.getTransitions())
		res.addTransition(pairLabel(label::labelFrom(AUTOMATON_FIRST), t.first.first), t.first.second, pairLabel(label::labelFrom(AUTOMATON_FIRST), t.second));

	for(const auto& t : second.getTransitions())
		res.addTransition(pairLabel(label::labelFrom(AUTOMATON_SECOND), t.first.first), t.first.second, pairLabel(label::labelFrom(AUTOMATON_SECOND), t.second));

	for(const auto& q : first.getFinalStates())
		res.addTransition(pairLabel(label::labelFrom(AUTOMATON_FIRST), q), pairLabel(label::labelFrom(AUTOMATON_SECOND), second.getInitialState()));

	return res;
}

auto AutomataConcatenationEpsilonTransitionDFA = AutomataConcatenationEpsilonTransition::RegistratorWrapper<automaton::EpsilonNFA, automaton::DFA>(AutomataConcatenationEpsilonTransition::getInstance(), AutomataConcatenationEpsilonTransition::concatenation);

automaton::EpsilonNFA AutomataConcatenationEpsilonTransition::concatenation(const automaton::NFA& first, const automaton::NFA& second) {
	automaton::EpsilonNFA res(pairLabel(label::labelFrom(AUTOMATON_FIRST), first.getInitialState()));

	for(const auto& symbol : first.getInputAlphabet())
		res.addInputSymbol(symbol);
	for(const auto& symbol : second.getInputAlphabet())
		res.addInputSymbol(symbol);

	for(const auto& q : first.getStates())
		res.addState(pairLabel(label::labelFrom(AUTOMATON_FIRST), q));
	for(const auto& q : second.getStates())
		res.addState(pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	for(const auto& q : second.getFinalStates())
		res.addFinalState(pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	for(const auto& t : first.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(pairLabel(label::labelFrom(AUTOMATON_FIRST), t.first.first), t.first.second, pairLabel(label::labelFrom(AUTOMATON_FIRST), q));

	for(const auto& t : second.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(pairLabel(label::labelFrom(AUTOMATON_SECOND), t.first.first), t.first.second, pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	for(const auto& q : first.getFinalStates())
		res.addTransition(pairLabel(label::labelFrom(AUTOMATON_FIRST), q), pairLabel(label::labelFrom(AUTOMATON_SECOND), second.getInitialState()));

	return res;
}

auto AutomataConcatenationEpsilonTransitionNFA = AutomataConcatenationEpsilonTransition::RegistratorWrapper<automaton::EpsilonNFA, automaton::NFA>(AutomataConcatenationEpsilonTransition::getInstance(), AutomataConcatenationEpsilonTransition::concatenation);

automaton::EpsilonNFA AutomataConcatenationEpsilonTransition::concatenation(const automaton::EpsilonNFA& first, const automaton::EpsilonNFA& second) {
	automaton::EpsilonNFA res(pairLabel(label::labelFrom(AUTOMATON_FIRST), first.getInitialState()));

	for(const auto& symbol : first.getInputAlphabet())
		res.addInputSymbol(symbol);
	for(const auto& symbol : second.getInputAlphabet())
		res.addInputSymbol(symbol);

	for(const auto& q : first.getStates())
		res.addState(pairLabel(label::labelFrom(AUTOMATON_FIRST), q));
	for(const auto& q : second.getStates())
		res.addState(pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	for(const auto& q : second.getFinalStates())
		res.addFinalState(pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	for(const auto& t : first.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(pairLabel(label::labelFrom(AUTOMATON_FIRST), t.first.first), t.first.second, pairLabel(label::labelFrom(AUTOMATON_FIRST), q));

	for(const auto& t : second.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(pairLabel(label::labelFrom(AUTOMATON_SECOND), t.first.first), t.first.second, pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	for(const auto& q : first.getFinalStates())
		res.addTransition(pairLabel(label::labelFrom(AUTOMATON_FIRST), q), pairLabel(label::labelFrom(AUTOMATON_SECOND), second.getInitialState()));

	return res;
}

auto AutomataConcatenationEpsilonTransitionEpsilonNFA = AutomataConcatenationEpsilonTransition::RegistratorWrapper<automaton::EpsilonNFA, automaton::EpsilonNFA>(AutomataConcatenationEpsilonTransition::getInstance(), AutomataConcatenationEpsilonTransition::concatenation);

} /* namespace transform */

} /* namespace automaton */
