/*
 * AutomataIntersectionCartesianProduct.cpp
 *
 *  Created on: 20. 11. 2014
 *	  Author: Tomas Pecka
 */

#include "AutomataIntersectionCartesianProduct.h"
#include <exception/AlibException.h>
#include "common/PairLabel.h"

#define AUTOMATON_FIRST  1
#define AUTOMATON_SECOND 2

namespace automaton {

namespace transform {

automaton::Automaton AutomataIntersectionCartesianProduct::intersection(const automaton::Automaton& first, const automaton::Automaton& second) {
	return getInstance().dispatch(first.getData(), second.getData());
}

automaton::DFA AutomataIntersectionCartesianProduct::intersection(const automaton::DFA& first, const automaton::DFA& second) {
	automaton::State q0(pairLabel(first.getInitialState(), second.getInitialState()));
	automaton::DFA res(q0);

	for(const auto& a : first.getInputAlphabet())
		res.addInputSymbol(a);
	for(const auto& a : second.getInputAlphabet())
		res.addInputSymbol(a);

	for(const auto& p : first.getStates())
		for(const auto& q : second.getStates())
			res.addState(automaton::State(pairLabel(p, q)));

	for(const auto& p : first.getFinalStates())
		for(const auto& q : second.getFinalStates())
			res.addFinalState(automaton::State(pairLabel(p, q)));

	for(const auto& state : res.getStates()) {
		const label::Label& label_p = static_cast<const label::LabelPairLabel&>(state.getName().getData()).getData().first;
		const label::Label& label_q = static_cast<const label::LabelPairLabel&>(state.getName().getData()).getData().second;

		for(const auto& tp : first.getTransitionsFromState(automaton::State(label_p)))
			for(const auto& tq : second.getTransitionsFromState(automaton::State(label_q)))
				if(tp.first.second == tq.first.second)
					res.addTransition(state, tp.first.second, automaton::State(pairLabel(tp.second, tq.second)));
	}

	return res;
}

auto AutomataIntersectionCartesianProductDFA = AutomataIntersectionCartesianProduct::RegistratorWrapper<automaton::DFA, automaton::DFA>(AutomataIntersectionCartesianProduct::getInstance(), AutomataIntersectionCartesianProduct::intersection);

automaton::NFA AutomataIntersectionCartesianProduct::intersection(const automaton::NFA& first, const automaton::NFA& second) {
	automaton::State q0(pairLabel(first.getInitialState(), second.getInitialState()));
	automaton::NFA res(q0);

	for(const auto& a : first.getInputAlphabet())
		res.addInputSymbol(a);
	for(const auto& a : second.getInputAlphabet())
		res.addInputSymbol(a);

	for(const auto& p : first.getStates())
		for(const auto& q : second.getStates())
			res.addState(automaton::State(pairLabel(p, q)));

	for(const auto& p : first.getFinalStates())
		for(const auto& q : second.getFinalStates())
			res.addFinalState(automaton::State(pairLabel(p, q)));

	for(const auto& state : res.getStates()) {
		const label::Label& label_p = static_cast<const label::LabelPairLabel&>(state.getName().getData()).getData().first;
		const label::Label& label_q = static_cast<const label::LabelPairLabel&>(state.getName().getData()).getData().second;

		for(const auto& tp : first.getTransitionsFromState(automaton::State(label_p)))
			for(const auto& tq : second.getTransitionsFromState(automaton::State(label_q)))
				if(tp.first.second == tq.first.second)
					for(const auto& p : tp.second)
						for(const auto& q : tq.second)
							res.addTransition(state, tp.first.second, automaton::State(pairLabel(p, q)));
	}

	return res;
}

auto AutomataIntersectionCartesianProductNFA = AutomataIntersectionCartesianProduct::RegistratorWrapper<automaton::NFA, automaton::NFA>(AutomataIntersectionCartesianProduct::getInstance(), AutomataIntersectionCartesianProduct::intersection);

} /* namespace transform */

} /* namespace automaton */
